#include <stdio.h>
#include <string.h>
#include <malloc.h>

/* This program should be compiled as a Windows or DOS4GW program, because
   otherwise it may erase any existing CW.ERR file.  There are some potential
   buffer overflow problems that are unlikely to be serious unless your map
   file has very large symbols. */

FILE * cwerr;
FILE * exemap;

struct {
	int len;
	unsigned int el[128];
	unsigned int base[128];
	char * msg[128];
} stack;

static char line[8192];
unsigned int start;
unsigned int length;
unsigned int eip;

void addAddress (unsigned int curr) {
int i;
	for (i=0; i < stack.len; i++) {
		if (start <= stack.el[i] && (stack.el[i] < start + length)) {
			if (curr <= stack.el[i] && stack.base[i] < curr) {
				stack.base[i] = curr;
			}
		}
	}
}

void addAddressDetail (unsigned int curr, char * msg) {
static char buff[128];
int i;

	for (i=0; msg[i] != '\0'; i++) {
		if (msg[i] == '\r' || msg[i] == '\n') {
			msg[i] = '\0';
			break;
		}
	}

	for (i=0; i < stack.len; i++) {
		if (stack.base[i] == curr) {
			sprintf (buff, "%08x = %s + 0x%x", curr, msg, stack.el[i] - stack.base[i]);
			stack.msg[i] = (char *)malloc (1 + strlen(buff));
			strcpy (stack.msg[i], buff);
		}
	}
}

void dumpStack () {
int i;
	for (i=0; i < stack.len; i++) {
		if (stack.msg[i] != NULL) {
			printf ("%s\n", stack.msg[i]);
		} else {
			printf ("%08x\n", stack.el[i]);
		}
	}
}

int main (int argc, char * argv[]) {
int state;
int i, c;
char * s;

	if (argc < 2) {
		printf ("%s  exemap [cw.err]", argv[0]);
		return -(__LINE__);
	}

	exemap = fopen (argv[1], "r");
	if (exemap == NULL) {
		printf ("unable to open %s\n", argv[1]);
		return -(__LINE__);
	}

	s = "CW.ERR";
	if (argc > 2) s = argv[2];
	cwerr = fopen (s, "r");

	if (cwerr == NULL) {
		printf ("unable to open %s\n", s);
		return -(__LINE__);
	}

	printf ("Ok, opened the files for reading\n");

	state = 0;
	s = (char *)&stack.el[0];
	while ((state < 4) && (fgets (line, 8191, cwerr) != NULL)) {
		switch (state) {
			case 0: if (strncmp ("EDI=", line, 4)) break;
					line[41] = '0';
					line[42] = 'x';
					line[51] = '\0';
					sscanf (line + 41, "%x", &eip);
					stack.el[0] = eip;
					s += sizeof (unsigned int);
					state = 1;
					break;

			case 1: if (strncmp ("Program Linear Load Address:", line, 28)) break;
					strcpy (line, line + 27);
					line[0] = '0';
					line[1] = 'x';
					sscanf (line, "%x", &start);
					/* start += 0x20; */
					state = 2;
					break;

			case 2: if (strncmp ("SS:ESP > ", line, 9)) break;
					state = 3;
					strcpy (line, line + 9);
			case 3: {
					char * l2;
					while (line[0] == ' ' || line[0] == '\t') {
						strcpy (line, line + 1);
					}
					if (line[0] == '\n' || line[0] == '\r') {
						state = 4;
						break;
					}
					l2 = line;
					do {
						int v;
						v = 0;
						for (i=0; (((l2[i]>='A') && (l2[i]<='F')) || ((l2[i]>='0') && (l2[i]<='9'))); i++) {
							v = 16 * v + l2[i];
							if (l2[i] >= 'A') v += 10 - 'A';
							else v -= '0';
						}
						if (i == 0) break;
						l2 += i+1;
						*s = v;
						s++;
					} while (1);
			}
		}
	}
	fclose (cwerr);
	stack.len = (s - (char *)&stack.el[0])/sizeof (unsigned int);

	printf ("start = %08x\n", start);
	printf ("Stack Length = %d\n", stack.len);

	for (i=0; i < 128; i++) {
		stack.base[i] = 0;
		stack.msg[i] = NULL;
	}

	printf ("------------------------------------------\n");

	c = 0;
	state = 0;
	while (fgets (line, 8191, exemap) != NULL) {

		if (state == 0 && strncmp ("_TEXT", line, 5) == 0) {
			line[67] = '0';
			line[68] = 'x';
			sscanf (line + 67, "%x", &length);
			printf ("length = %08x (max = %08x)\n", length, start+length);
			length -= 0x10;
        }

        if (0 == strncmp ("Module:", line, 7)) {
			state = 1;
		} else if (state == 1) {
			if (0 == strncmp ("0001:", line, 5)) {
				unsigned int addr;
				line[3] = '0';
				line[4] = 'x';
				line[13] = '\0';
				sscanf (line+3, "%x", &addr);
				addAddress (start+addr);
				/* addAddressDetail (addr, line+15);*/
				c++;
			}
		}
	}

	exemap = freopen (argv[1], "r", exemap);

	state = 0;
	while (fgets (line, 8191, exemap) != NULL) {
		if (0 == strncmp ("Module:", line, 7)) {
			state = 1;
		} else if (state == 1) {
			if (0 == strncmp ("0001:", line, 5)) {
				unsigned int addr;
				line[3] = '0';
				line[4] = 'x';
				line[13] = '\0';
				sscanf (line+3, "%x", &addr);
				addAddressDetail (start+addr, line+15);
				c--;
				if (c == 0) break;
			}
		}
	}

	printf ("Done\n"); fflush (stdout);

	dumpStack ();
	fcloseall ();

	return 0;
}
